@@ -11,6 +11,7 @@ from typing import Tuple
 
 import requests
 import yaml
+from more_itertools import take
 from pystac import STACValidationError
 from pystac_client import Client
 from shapely.geometry import shape
@@ -207,12 +208,8 @@ def validate_api(
     if not root_body.get("links"):
         errors.append("/ : 'links' field must be defined and non-empty.")
 
-    if "core" in conformance_classes and not any(
-        cc_core_regex.fullmatch(x) for x in conforms_to
-    ):
-        errors.append(
-            "/: Core configured for validation, but not contained in 'conformsTo'"
-        )
+    if not any(cc_core_regex.fullmatch(x) for x in conforms_to):
+        errors.append("/: STAC API - Core not contained in 'conformsTo'")
 
     if "browseable" in conformance_classes and not any(
         cc_browseable_regex.fullmatch(x) for x in conforms_to
@@ -265,30 +262,43 @@ def validate_api(
     if errors:
         return warnings, errors
 
-    print("STAC API - Core conformance class found.")
+    print("Validating STAC API - Core conformance class.")
     validate_core(root_body, warnings, errors)
 
     if "browseable" in conformance_classes:
-        print("STAC API - Browseable conformance class found.")
+        print("Validating STAC API - Browseable conformance class.")
         validate_browseable(root_body, warnings, errors)
+    else:
+        print("Skipping STAC API - Browseable conformance class.")
 
     if "children" in conformance_classes:
-        print("STAC API - Children conformance class found.")
+        print("Validating STAC API - Children conformance class.")
         validate_children(root_body, warnings, errors)
+    else:
+        print("Skipping STAC API - Children conformance class.")
 
     if "collections" in conformance_classes:
-        print("STAC API - Collections conformance class found.")
+        print("Validating STAC API - Collections conformance class.")
         validate_collections(root_body, collection, warnings, errors)  # type:ignore
+    else:
+        print("Skipping STAC API - Collections conformance class.")
 
     if "features" in conformance_classes:
-        print("STAC API - Features conformance class found.")
+        print("Validating STAC API - Features conformance class.")
         validate_collections(root_body, collection, warnings, errors)  # type:ignore
         validate_features(
-            root_body, conforms_to, collection, warnings, errors  # type:ignore
+            root_body,
+            conforms_to,
+            collection,
+            geometry,
+            warnings,
+            errors,
         )
+    else:
+        print("Skipping STAC API - Features conformance class.")
 
     if "item-search" in conformance_classes:
-        print("STAC API - Item Search conformance class found.")
+        print("Validating STAC API - Item Search conformance class.")
         validate_item_search(
             root_url=root_url,
             root_body=root_body,
@@ -300,6 +310,8 @@ def validate_api(
             geometry=geometry,  # type:ignore
             conformance_classes=conformance_classes,
         )
+    else:
+        print("Skipping STAC API - Item Search conformance class.")
 
     if not errors:
         try:
@@ -314,8 +326,8 @@ def validate_api(
 
 
 def link_by_rel(
-    links: Optional[List[Dict[str, str]]], rel: str
-) -> Optional[Dict[str, str]]:
+    links: Optional[List[Dict[str, Any]]], rel: str
+) -> Optional[Dict[str, Any]]:
     if not links:
         return None
     else:
@@ -327,6 +339,9 @@ def validate_core(
 ) -> None:
     links = root_body.get("links")
 
+    if links is None:
+        errors.append("/ : 'links' attribute missing")
+
     if not (root := link_by_rel(links, "root")):
         errors.append("/ : Link[rel=root] must exist")
     else:
@@ -446,12 +461,21 @@ def validate_collections(
 def validate_features(
     root_body: Dict[str, Any],
     conforms_to: List[str],
-    collection: str,
+    collection: Optional[str],
+    geometry: Optional[str],
     warnings: List[str],
     errors: List[str],
 ) -> None:
     print("WARNING: Features validation is not yet fully implemented.")
 
+    if not geometry:
+        errors.append("Geometry parameter required for running Features validations.")
+        return
+
+    if not collection:
+        errors.append("Collection parameter required for running Features validations.")
+        return
+
     if conforms_to and (
         req_ccs := [
             x
@@ -505,7 +529,7 @@ def validate_features(
             conformance_json = r_conformance.json()
             if not (
                 set(root_body.get("conformsTo", []))
-                == set(conformance_json["conformsTo"])
+                == set(conformance_json.get("conformsTo", []))
             ):
                 warnings.append(
                     "Landing Page conforms to and conformance conformsTo must be the same",
@@ -515,24 +539,42 @@ def validate_features(
                 f"service-desc ({conformance}): must return JSON, instead got non-JSON text"
             )
 
-        # this is hard to figure out, since it's likely a mistake, but most apis can't undo it for
-        # backwards-compat reasons
-        if not (link_by_rel(root_links, "collections") is None):
-            warnings.append(
-                "/ Link[rel=collections] is a non-standard relation. Use Link[rel=data instead]"
+    # this is hard to figure out, since it's likely a mistake, but most apis can't undo it for
+    # backwards-compat reasons
+    if not (link_by_rel(root_links, "collections") is None):
+        warnings.append(
+            "/ Link[rel=collections] is a non-standard relation. Use Link[rel=data instead]"
+        )
+
+    # todo: validate items exists in the collection
+
+    if not (collections_url := link_by_rel(root_links, "data")):
+        errors.append("/: Link[rel=data] must href /collections")
+    else:
+        item_url = f"{collections_url['href']}/{collection}/items/non-existent-item"
+        r = requests.get(item_url)
+        if r.status_code != 404:
+            errors.append(
+                f"[Features] GET {item_url} (non-existent item) returned status code {r.status_code} instead of 404"
             )
 
-        # todo: validate items exists
-
-        if not (collections_url := link_by_rel(root_links, "data")):
-            errors.append("/: Link[rel=data] must href /collections")
+    if not (collections_url := link_by_rel(root_links, "data")):
+        errors.append(
+            "/: Link[rel=data] must href /collections, cannot run pagination test"
+        )
+    else:
+        if not (self_link := link_by_rel(root_links, "self")):
+            errors.append("/: Link[rel=self] missing")
         else:
-            item_url = f"{collections_url['href']}/{collection}/items/non-existent-item"
-            r = requests.get(item_url)
-            if r.status_code != 404:
-                errors.append(
-                    f"[Features] GET {item_url} (non-existent item) returned status code {r.status_code} instead of 404"
-                )
+            validate_item_pagination(
+                root_url=self_link.get("href", ""),
+                search_url=f"{collections_url['href']}/{collection}/items",
+                collection=None,
+                geometry=geometry,
+                post=False,
+                errors=errors,
+                use_pystac_client=False,
+            )
 
     # if any(cc_features_fields_regex.fullmatch(x) for x in conforms_to):
     #     print("STAC API - Features - Fields extension conformance class found.")
@@ -550,12 +592,14 @@ def validate_features(
     #     print("STAC API - Features - Filter extension conformance class found.")
 
     if any(cc_features_filter_regex.fullmatch(x) for x in conforms_to):
-        print("STAC API - Features - Filter extension conformance class found.")
+        print("STAC API - Features - Filter Extension conformance class found.")
         validate_features_filter(
             root_body=root_body,
             collection=collection,
             errors=errors,
         )
+    else:
+        print("Skipping STAC API - Features - Filter Extension conformance class.")
 
 
 def validate_item_search(
@@ -620,6 +664,15 @@ def validate_item_search(
         geometry=geometry,
     )
 
+    validate_item_pagination(
+        root_url=root_url,
+        search_url=search_url,
+        collection=collection,
+        geometry=geometry,
+        post=post,
+        errors=errors,
+    )
+
     # if any(cc_item_search_fields_regex.fullmatch(x) for x in conforms_to):
     #     print("STAC API - Item Search - Fields extension conformance class found.")
     #
@@ -649,8 +702,7 @@ def validate_item_search(
         or x.endswith("item-search#filter:filter")
         for x in conforms_to
     ):
-        print("STAC API - Item Search - Filter extension conformance class found.")
-
+        print("Validating STAC API - Item Search - Filter Extension conformance class.")
         validate_item_search_filter(
             root_url=root_url,
             root_body=root_body,
@@ -659,6 +711,8 @@ def validate_item_search(
             warnings=warnings,
             errors=errors,
         )
+    else:
+        print("Skipping STAC API - Item Search - Filter Extension conformance class.")
 
 
 def validate_filter_queryables(
@@ -753,7 +807,7 @@ def validate_item_search_filter(
         )
     ):
         errors.append(
-            "[Item Search Filter Ext] / : 'http://www.opengis.net/def/rel/ogc/1.0/queryables' (Queryables) link relation missing"
+            "[Item Search - Filter Ext] / : 'http://www.opengis.net/def/rel/ogc/1.0/queryables' (Queryables) link relation missing"
         )
 
     validate_filter_queryables(
@@ -763,7 +817,7 @@ def validate_item_search_filter(
         errors=errors,
     )
 
-    conforms_to = root_body["conformsTo"]
+    conforms_to = root_body.get("conformsTo", [])
 
     cql2_text_supported = (
         "http://www.opengis.net/spec/cql2/1.0/conf/cql2-text" in conforms_to
@@ -773,7 +827,13 @@ def validate_item_search_filter(
     )
 
     if cql2_text_supported:
-        print("CQL2 - CQL2-Text conformance class found.")
+        print(
+            "Validating STAC API - Item Search - Filter Extension - CQL2-Text conformance class."
+        )
+    else:
+        print(
+            "Skipping STAC API - Item Search - Filter Extension - CQL2-Text conformance class."
+        )
 
     cql2_json_supported = (
         "http://www.opengis.net/spec/cql2/1.0/conf/cql2-json" in conforms_to
@@ -783,7 +843,13 @@ def validate_item_search_filter(
     )
 
     if cql2_json_supported:
-        print("CQL2 - CQL2-JSON conformance class found.")
+        print(
+            "Validating STAC API - Item Search - Filter Extension - CQL2-JSON conformance class."
+        )
+    else:
+        print(
+            "Skipping STAC API - Item Search - Filter Extension - CQL2-JSON conformance class."
+        )
 
     basic_cql2_supported = (
         "http://www.opengis.net/spec/cql2/1.0/conf/basic-cql2" in conforms_to
@@ -793,14 +859,27 @@ def validate_item_search_filter(
     )
 
     if basic_cql2_supported:
-        print("CQL2 - Basic CQL2 conformance class found.")
+        print(
+            "Validating STAC API - Item Search - Filter Extension - Basic CQL2 conformance class."
+        )
+    else:
+        print(
+            "Skipping STAC API - Item Search - Filter Extension - Basic CQL2 conformance class."
+        )
 
     advanced_comparison_operators_supported = (
         "http://www.opengis.net/spec/cql2/1.0/conf/advanced-comparison-operators"
         in conforms_to
     )
+
     if advanced_comparison_operators_supported:
-        print("CQL2 - Advanced Comparison Operators conformance class found.")
+        print(
+            "Validating STAC API - Item Search - Filter Extension - Advanced Comparison Operators conformance class."
+        )
+    else:
+        print(
+            "Skipping STAC API - Item Search - Filter Extension - Advanced Comparison Operators conformance class."
+        )
 
     basic_spatial_operators_supported = (
         "http://www.opengis.net/spec/cql2/1.0/conf/basic-spatial-operators"
@@ -808,14 +887,26 @@ def validate_item_search_filter(
     )
 
     if basic_spatial_operators_supported:
-        print("CQL2 - Basic Spatial Operators conformance class found.")
+        print(
+            "Validating STAC API - Item Search - Filter Extension - Basic Spatial Operators conformance class."
+        )
+    else:
+        print(
+            "Skipping STAC API - Item Search - Filter Extension - Basic Spatial Operators conformance class."
+        )
 
     temporal_operators_supported = (
         "http://www.opengis.net/spec/cql2/1.0/conf/temporal-operators" in conforms_to
     )
 
     if temporal_operators_supported:
-        print("CQL2 - Temporal Operators conformance class found.")
+        print(
+            "Validating STAC API - Item Search - Filter Extension - Temporal Operators conformance class."
+        )
+    else:
+        print(
+            "Skipping STAC API - Item Search - Filter Extension - Temporal Operators conformance class."
+        )
 
     # todo: validate these
     # Spatial Operators: http://www.opengis.net/spec/cql2/1.0/conf/spatial-operators
@@ -906,7 +997,7 @@ def validate_item_search_filter(
         )
         if not (r.status_code == 200):
             errors.append(
-                f"[Item Search Filter Ext] GET {f_text} returned status code {r.status_code}"
+                f"[Item Search - Filter Ext] GET {f_text} returned status code {r.status_code}"
             )
 
     for f_json in filter_jsons:
@@ -915,7 +1006,7 @@ def validate_item_search_filter(
         )
         if r.status_code != 200:
             errors.append(
-                f"[Item Search Filter Ext] POST {f_json} returned status code {r.status_code}"
+                f"[Item Search - Filter Ext] POST {f_json} returned status code {r.status_code}"
             )
 
 
@@ -985,6 +1076,166 @@ def validate_item_search_bbox_xor_intersects(
             )
 
 
+def validate_item_pagination(
+    root_url: str,
+    search_url: str,
+    collection: Optional[str],
+    geometry: str,
+    post: bool,
+    errors: List[str],
+    use_pystac_client: bool = True,
+) -> None:
+    url = f"{search_url}?limit=1"
+    if collection is not None:
+        url = f"{url}&collections={collection}"
+
+    r = requests.get(url)
+    if not r.status_code == 200:
+        errors.append(
+            "STAC API - Item Search GET pagination get failed for initial request"
+        )
+    else:
+        try:
+            first_body = r.json()
+            if link := link_by_rel(first_body.get("links"), "next"):
+                if (method := link.get("method")) and method != "GET":
+                    errors.append(
+                        f"STAC API - Item Search GET pagination first request 'next' link relation has method {method} instead of 'GET'"
+                    )
+
+                next_url = link.get("href")
+                if next_url is None:
+                    errors.append(
+                        "STAC API - Item Search GET pagination first request 'next' link relation missing href"
+                    )
+                else:
+                    if url == next_url:
+                        errors.append(
+                            "STAC API - Item Search GET pagination next href same as first url"
+                        )
+
+                    r = requests.get(next_url)
+                    if not r.status_code == 200:
+                        errors.append(
+                            f"STAC API - Item Search GET pagination get failed for next url {next_url}"
+                        )
+            else:
+                errors.append(
+                    "STAC API - Item Search GET pagination first request had no 'next' link relation"
+                )
+
+        except json.decoder.JSONDecodeError:
+            errors.append(
+                f"STAC API - Item Search GET pagination response failed {url}"
+            )
+
+    max_items = 100
+
+    # todo: how to paginate over items, not just search?
+
+    if use_pystac_client:
+        client = Client.open(root_url)
+        search = client.search(
+            method="GET", collections=[collection], max_items=max_items, limit=5
+        )
+
+        items = list(search.items_as_dicts())
+
+        if len(items) > max_items:
+            errors.append(
+                "STAC API - Item Search GET pagination - more than max items returned from paginating"
+            )
+
+        if len(items) > len({item["id"] for item in items}):
+            errors.append(
+                "STAC API - Item Search GET pagination - duplicate items returned from paginating items"
+            )
+
+    # GET paging has a problem with intersects https://github.com/stac-utils/pystac-client/issues/335
+    # search = client.search(method="GET", collections=[collection], intersects=geometry)
+    # if len(list(take(20000, search.items_as_dicts()))) == 20000:
+    #     errors.append(
+    #         f"STAC API - Item Search GET pagination - paged through 20,000 results. This could mean the last page "
+    #         f"of results references itself, or your collection and geometry combination has too many results."
+    #     )
+
+    if post:
+        initial_json_body = {"limit": 1, "collections": [collection]}
+        r = requests.post(search_url, json=initial_json_body)
+        if not r.status_code == 200:
+            errors.append(
+                "STAC API - Item Search POST pagination get failed for initial request"
+            )
+        else:
+            try:
+                first_body = r.json()
+                if link := link_by_rel(first_body.get("links"), "next"):
+                    if (method := link.get("method")) and method != "POST":
+                        errors.append(
+                            f"STAC API - Item Search POST pagination first request 'next' link relation has method {method} instead of 'POST'"
+                        )
+
+                    next_url = link.get("href")
+                    if next_url is None:
+                        errors.append(
+                            "STAC API - Item Search POST pagination first request 'next' link relation missing href"
+                        )
+                    else:
+                        if url == next_url:
+                            errors.append(
+                                "STAC API - Item Search POST pagination next href same as first url"
+                            )
+
+                        next_body: Dict[str, Any] = link.get("body", {})
+                        if not link.get("merge", False):
+                            second_json_body = next_body
+                        else:
+                            second_json_body = initial_json_body
+                            second_json_body.update(next_body)
+
+                        r = requests.post(next_url, json=second_json_body)
+                        if not r.status_code == 200:
+                            errors.append(
+                                f"STAC API - Item Search POST pagination get failed for next url {next_url} with body {second_json_body}"
+                            )
+                        else:
+                            r.json()
+                else:
+                    errors.append(
+                        "STAC API - Item Search POST pagination first request had no 'next' link relation"
+                    )
+
+            except json.decoder.JSONDecodeError:
+                errors.append("STAC API - Item Search POST pagination response failed")
+
+        max_items = 100
+        client = Client.open(root_url)
+        search = client.search(
+            method="POST", collections=[collection], max_items=max_items, limit=5
+        )
+
+        items = list(search.items_as_dicts())
+
+        if len(items) > max_items:
+            errors.append(
+                "STAC API - Item Search POST pagination - more than max items returned from paginating"
+            )
+
+        if len(items) > len({item["id"] for item in items}):
+            errors.append(
+                "STAC API - Item Search POST pagination - duplicate items returned from paginating items"
+            )
+
+        search = client.search(
+            method="POST", collections=[collection], intersects=geometry
+        )
+        if len(list(take(20000, search.items_as_dicts()))) == 20000:
+            errors.append(
+                "STAC API - Item Search POST pagination - paged through 20,000 results. This could mean the last page "
+                "of results references itself, or your collection and geometry combination has too many results."
+            )
+
+
 def validate_item_search_intersects(
     search_url: str, collection: str, post: bool, errors: List[str], geometry: str
 ) -> None:
